<?php

namespace Drupal\media_migration\Plugin\migrate;

use Drupal\Component\Plugin\Derivative\DeriverBase;
use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\Discovery\ContainerDeriverInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\media\MediaTypeInterface;
use Drupal\migrate\Exception\RequirementsException;
use Drupal\migrate\Plugin\MigratePluginManagerInterface;
use Drupal\migrate\Plugin\MigrationDeriverTrait;
use Drupal\migrate_drupal\FieldDiscoveryInterface;
use Drupal\migrate_drupal\Plugin\MigrateFieldPluginManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Deriver for file entities.
 */
class D7FileEntityItemDeriver extends DeriverBase implements ContainerDeriverInterface {

  use MigrationDeriverTrait;
  use StringTranslationTrait;

  /**
   * The base plugin ID this derivative is for.
   *
   * @var string
   */
  protected $basePluginId;

  /**
   * Already-instantiated field plugins, keyed by ID.
   *
   * @var \Drupal\migrate_drupal\Plugin\MigrateFieldInterface[]
   */
  protected $fieldPluginCache;

  /**
   * The field plugin manager.
   *
   * @var \Drupal\migrate_drupal\Plugin\MigrateFieldPluginManagerInterface
   */
  protected $fieldPluginManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The migration field discovery service.
   *
   * @var \Drupal\migrate_drupal\FieldDiscoveryInterface
   */
  protected $fieldDiscovery;

  /**
   * The ID of the null coalesce migrate process plugin to use.
   *
   * The 'null_coalesce' plugin was introduced in Drupal core 8.8.x, but we
   * need it for being able to properly handle image media alt and title
   * migrations. So we define our own, equal plugin; but we use it only if the
   * 'null_coalesce' from core is not available.
   *
   * @var string
   *
   * @todo Remove when Drupal core 8.7.x security support ends.
   */
  protected $nullCoalesceMigrateProcessPluginId;

  /**
   * D7FileEntityItemDeriver constructor.
   *
   * @param string $base_plugin_id
   *   The base plugin ID for the plugin ID.
   * @param \Drupal\migrate_drupal\Plugin\MigrateFieldPluginManagerInterface $field_manager
   *   The field plugin manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\migrate_drupal\FieldDiscoveryInterface $field_discovery
   *   The migration field discovery service.
   * @param \Drupal\migrate\Plugin\MigratePluginManagerInterface $process_plugin_manager
   *   The process migration plugin manager.
   */
  public function __construct($base_plugin_id, MigrateFieldPluginManagerInterface $field_manager, EntityTypeManagerInterface $entity_type_manager, FieldDiscoveryInterface $field_discovery, MigratePluginManagerInterface $process_plugin_manager) {
    $this->basePluginId = $base_plugin_id;
    $this->fieldPluginManager = $field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->fieldDiscovery = $field_discovery;
    $this->nullCoalesceMigrateProcessPluginId = $process_plugin_manager->hasDefinition('null_coalesce') ?
      'null_coalesce' :
      'media_migration_null_coalesce';
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, $base_plugin_id) {
    return new static(
      $base_plugin_id,
      $container->get('plugin.manager.migrate.field'),
      $container->get('entity_type.manager'),
      $container->get('migrate_drupal.field_discovery'),
      $container->get('plugin.manager.migrate.process')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivativeDefinitions($base_plugin_definition) {
    $types = static::getSourcePlugin('d7_file_entity_type');

    try {
      $types->checkRequirements();
    }
    catch (RequirementsException $e) {
      return $this->derivatives;
    }

    try {
      foreach ($types as $row) {
        /** @var \Drupal\migrate\Row $row */
        $derivative_definition = $base_plugin_definition;
        $bundle_name = $row->getSourceProperty('type');
        $derivative_id = $bundle_name;

        $media_type = NULL;
        try {
          $media_type = $this->entityTypeManager->getStorage('media_type')
            ->load($bundle_name);
        }
        // Catch plugin not found and invalid plugin definition exceptions.
        catch (PluginNotFoundException $e) {
          continue;
        }
        catch (InvalidPluginDefinitionException $e) {
          continue;
        }

        // If this media type does not exist yet or the variable not a media
        // type, continue to the next one.
        if (empty($media_type) || !($media_type instanceof MediaTypeInterface)) {
          continue;
        }

        $source_field_name = $media_type->getSource()->getConfiguration()['source_field'];

        // Create the migration derivative.
        $derivative_definition['source']['type'] = $bundle_name;
        $derivative_definition['label'] = $this->t('@label (@type)', [
          '@label' => $base_plugin_definition['label'],
          '@type' => $media_type->label(),
        ]);

        // Media Migration uses this derivative class for multiple migrations,
        // but the actual media entity migration derivatives need special
        // treatment.
        if ($base_plugin_definition['source']['plugin'] === 'd7_file_entity_item') {
          $derivative_definition['destination']['bundle'] = $bundle_name;
          $derivative_definition['process'][$source_field_name . '/target_id'] = 'fid';

          // Map width and height for image entities.
          if ($bundle_name == 'image') {
            $derivative_definition['process'][$source_field_name . '/width'] = 'width';
            $derivative_definition['process'][$source_field_name . '/height'] = 'height';
            $derivative_definition['process']['thumbnail/target_id'] = 'fid';
            $derivative_definition['process']['thumbnail/width'] = 'width';
            $derivative_definition['process']['thumbnail/height'] = 'height';
          }

          /** @var \Drupal\migrate\Plugin\Migration $migration */
          $migration = \Drupal::service('plugin.manager.migration')
            ->createStubMigration($derivative_definition);
          $this->fieldDiscovery->addBundleFieldProcesses($migration, 'file', $bundle_name);
          $derivative_definition = $migration->getPluginDefinition();

          if ($media_type->getSource()->getPluginId() === 'image') {
            // These property fields only exist when the file_entity module is
            // installed on the source site.
            $property_fields = [
              'field_file_image_alt_text' => 'alt',
              'field_file_image_title_text' => 'title',
            ];

            foreach ($property_fields as $field_name => $property) {
              $derivative_definition['process']["{$property}_from_media"] = [
                [
                  'plugin' => 'extract',
                  'source' => $field_name,
                  'index' => ['0', 'value'],
                  // It seems to be impossible to set 'NULL' as default value.
                  // Using ['default => NULL'] is equal with not setting the key
                  // at all (so every image media migraqqqtion that should be
                  // migrated from an image field will be skipped).
                  // Setting the default to a predefined constant that equals to
                  // NULL doesn't work either. For example, if we set 'default'
                  // to 'constants/alt_and_title_default', then every image
                  // that's alt is empty will be migrated with the
                  // 'constants/alt_and_title_default' string set as alt
                  // property. If we don't set 'default' at all, we will get a
                  // 'Notice: Undefined index: default_value' exception for
                  // empty source values.
                  // @todo Revisit after https://www.drupal.org/node/3133516 is
                  //   fixed.
                  'default' => '',
                ],
                [
                  'plugin' => 'default_value',
                  'default_value' => NULL,
                ],
              ];

              $property_process = [
                [
                  'plugin' => $this->nullCoalesceMigrateProcessPluginId,
                  'source' => [
                    "image_field_$property",
                    "@{$property}_from_media",
                  ],
                  'default_value' => NULL,
                ],
              ];

              $derivative_definition['process'][$source_field_name . '/' . $property] =
              $derivative_definition['process']['thumbnail/' . $property] =
                $property_process;
            }
          }
        }

        // Post-process migration dependencies: instead of depending on
        // migrations based on their base plugin ID, it is better to use the
        // corresponding derivatives where possible.
        $derived_migration_ids = [
          'd7_file_entity',
          'd7_file_entity_type',
          'd7_media_source_field_config',
        ];
        foreach ($derived_migration_ids as $derived_migration_base_id) {
          $required_dependencies = !empty($derivative_definition['migration_dependencies']['required'])
            ? $derivative_definition['migration_dependencies']['required']
            : [];
          $dependency_key = array_search($derived_migration_base_id, $required_dependencies, TRUE);
          if ($dependency_key !== FALSE) {
            $derivative_definition['migration_dependencies']['required'][$dependency_key] .= ":$bundle_name";
          }
        }

        $this->derivatives[$derivative_id] = $derivative_definition;
      }
    }
    catch (DatabaseExceptionWrapper $e) {
      // Once we begin iterating the source plugin it is possible that the
      // source tables will not exist. This can happen when the
      // MigrationPluginManager gathers up the migration definitions but we do
      // not actually have a Drupal 7 source database.
    }
    return $this->derivatives;
  }

}
