<?php

namespace Drupal\media_migration\Plugin\migrate\process;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\migrate\MigrateExecutableInterface;
use Drupal\migrate\ProcessPluginBase;
use Drupal\migrate\Row;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure field instance settings for file entity fields.
 *
 * @MigrateProcessPlugin(
 *   id = "file_entity_field_instance_settings"
 * )
 */
class FileEntityFieldInstanceSettings extends ProcessPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The available media type IDs.
   *
   * @var string[]
   */
  protected $mediaTypeIds;

  /**
   * Constructs a FileEntityFieldInstanceSettings instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->mediaTypeIds = $entity_type_manager->hasDefinition('media_type')
      ? array_keys($entity_type_manager->getStorage('media_type')->loadMultiple())
      : [];
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function transform($value, MigrateExecutableInterface $migrate_executable, Row $row, $destination_property) {
    $type = $row->getSourceProperty('type');

    if ($type == 'file_entity') {
      $widget_settings = $row->getSourceProperty('widget');
      $target_bundles = !empty($widget_settings['settings']['allowed_types'])
        ? array_filter($widget_settings['settings']['allowed_types'])
        : [];

      // We have to map "file" type to "document" for sites installed with
      // Drupal core version prior 8.8.0.
      $document_type_is_missing = !in_array('document', $this->mediaTypeIds, TRUE);
      $file_type_is_available = in_array('file', $this->mediaTypeIds, TRUE);
      if ($document_type_is_missing && $file_type_is_available) {
        $document_key = array_search('document', $target_bundles);
        if ($document_key !== FALSE) {
          unset($target_bundles[$document_key]);
          $target_bundles['file'] = 'file';
        }
      }

      // In Drupal 7, when no media types are explicitly enabled on this field,
      // that means that every media type is allowed. In Drupal 8|9, this
      // feature is not available anymore: "target_bundles" cannot be empty.
      $value['handler_settings']['target_bundles'] = !empty($target_bundles)
        ? $target_bundles
        : array_combine($this->mediaTypeIds, $this->mediaTypeIds);
    }
    return $value;
  }

}
