<?php

namespace Drupal\media_migration\Plugin\migrate\source\d7;

use Drupal\migrate\Row;
use Drupal\migrate_drupal\Plugin\migrate\source\d7\FieldableEntity;

/**
 * File Entity Item source plugin.
 *
 * Available configuration keys:
 * - type: (optional) If supplied, this will only return fields
 *   of that particular type.
 *
 * @MigrateSource(
 *   id = "d7_file_entity_item",
 *   source_module = "file_entity",
 * )
 */
class FileEntityItem extends FieldableEntity {

  use MediaFieldQueryTrait;

  /**
   * {@inheritdoc}
   */
  public function query() {
    $query = $this->select('file_managed', 'f')
      ->fields('f')
      ->orderBy('f.timestamp');

    // Filter by type, if configured.
    if (isset($this->configuration['type'])) {
      $op = is_array($this->configuration['type']) ? 'IN' : '=';
      $query->condition('type', $this->configuration['type'], $op);
    }

    // Filter by URI prefix if specified. Default to 'public://'.
    if (isset($this->configuration['uri_prefix'])) {
      $query->condition('uri', $this->configuration['uri_prefix'] . '%', 'LIKE');
    }
    else {
      $query->condition('uri', 'public://%', 'LIKE');
    }

    return $query;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareRow(Row $row) {
    $media_bundle = $row->getSourceProperty('type');
    // Get Field API field values.
    foreach (array_keys($this->getFields('file', $media_bundle)) as $field) {
      $fid = $row->getSourceProperty('fid');
      $row->setSourceProperty($field, $this->getFieldValues('file', $field, $fid));
    }

    if ($media_bundle === 'image') {
      // Add width and height source properties for image entities. These
      // properties only exist when the file_entity module is installed on the
      // source site.
      $width_and_height_statement = $this->getDatabase()->select('file_metadata', 'fm')
        ->fields('fm', ['name', 'value'])
        ->condition('fm.fid', $row->getSourceProperty('fid'))
        ->condition('fm.name', ['width', 'height'], 'IN')
        ->execute();
      foreach ($width_and_height_statement as $result_row) {
        $row->setSourceProperty($result_row->name, unserialize($result_row->value));
      }

      // Add alt and title properties from image type fields where the current
      // image is the file value.
      $image_field_names = $this->getImageFieldNames();

      $alt_from_field_content = NULL;
      $title_from_field_content = NULL;
      foreach ($image_field_names as $field_name) {
        $alt_and_title_from_field_content_query = $this->getDatabase()->select("field_data_$field_name", $field_name);
        $alt_and_title_from_field_content_query->addField($field_name, "{$field_name}_alt", 'alt');
        $alt_and_title_from_field_content_query->addField($field_name, "{$field_name}_title", 'title');
        $alt_and_title_from_field_content_query->condition("$field_name.{$field_name}_fid", $row->getSourceProperty('fid'));
        $alt_and_title_from_field_content_query->orderBy("$field_name.revision_id", 'DESC');
        $alt_and_title_from_field_content = $alt_and_title_from_field_content_query->execute()->fetchAll();

        if (count($alt_and_title_from_field_content) > 0) {
          // We may find more than one alt-title props in other fields – we use
          // the first row.
          if (!$alt_from_field_content && !$title_from_field_content) {
            $alt_from_field_content = $alt_from_field_content ?? $alt_and_title_from_field_content[0]->alt;
            $title_from_field_content = $title_from_field_content ?? $alt_and_title_from_field_content[0]->title;
          }
        }
      }

      $row->setSourceProperty('image_field_alt', $alt_from_field_content);
      $row->setSourceProperty('image_field_title', $title_from_field_content);
    }

    return parent::prepareRow($row);
  }

  /**
   * {@inheritdoc}
   */
  public function fields() {
    // Fields provided by file_admin module are only included here for developer
    // convenience so that they can be adjusted by altering the generated
    // migration plugins.
    $fields = [
      'fid' => $this->t('The file identifier'),
      'uid' => $this->t('The user identifier'),
      'filename' => $this->t('The file name'),
      'uri' => $this->t('The URI of the file'),
      'filemime' => $this->t('The file mimetype'),
      'filesize' => $this->t('The file size'),
      'status' => $this->t('The file status'),
      'timestamp' => $this->t('The time that the file was added'),
      'type' => $this->t('The file type'),
      'created' => $this->t('The created timestamp - (if file_admin module is present in Drupal 7)'),
      'published' => $this->t('The published timestamp - (if file_admin module is present in Drupal 7)'),
      'promote' => $this->t('The promoted flag - (if file_admin module is present in Drupal 7)'),
      'sticky' => $this->t('The sticky flag - (if file_admin module is present in Drupal 7)'),
      'vid' => $this->t('The vid'),
    ];

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function getIds() {
    $ids['fid']['type'] = 'integer';
    return $ids;
  }

}
