<?php

namespace Drupal\media_migration\Plugin\migrate\source\d7;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StreamWrapper\StreamWrapperInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\media\MediaTypeInterface;
use Drupal\media\Plugin\media\Source\File;
use Drupal\media\Plugin\media\Source\Image;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate\Row;
use Drupal\migrate_drupal\Plugin\migrate\source\DrupalSqlBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Drupal 7 media source field instances source based on source database.
 *
 * @MigrateSource(
 *   id = "d7_media_source_field_instance",
 *   source_module = "file_entity"
 * )
 */
class MediaSourceFieldInstance extends DrupalSqlBase {

  use MediaFieldQueryTrait;

  /**
   * The map of the alt and title properties and their corresponding field name.
   *
   * These property fields only exist when the file_entity module is installed
   * on the source site.
   *
   * @var string[]
   */
  protected $fieldNamePropertyMap = [
    'field_file_image_alt_text' => 'alt_field',
    'field_file_image_title_text' => 'title_field',
  ];

  /**
   * The list of the visible, writable and readable streams.
   *
   * @var string[]
   */
  protected $publicStreams;

  /**
   * The media type's storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $mediaTypeStorage;

  /**
   * Constructs a MediaSourceFieldInstance instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The current migration.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $stream_wrapper_manager
   *   The stream wrapper manager.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *   Thrown when the media_type entity type doesn't exist.
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   *   Thrown when the media_type storage handler couldn't be loaded.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, MigrationInterface $migration, StateInterface $state, EntityTypeManagerInterface $entity_type_manager, StreamWrapperManagerInterface $stream_wrapper_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $migration, $state, $entity_type_manager);
    // For the file migration, we need writable streams. This also means that
    // these streams are readable as well. On the other hand, only visible
    // streams can be configured for file|image fields on Drupal 8|9, and
    // Media Migration wants to respect this decision.
    // @see \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface::WRITE_VISIBLE
    $this->publicStreams = array_keys($stream_wrapper_manager->getNames(StreamWrapperInterface::WRITE_VISIBLE));
    $this->mediaTypeStorage = $entity_type_manager->getStorage('media_type');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition, MigrationInterface $migration = NULL) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $migration,
      $container->get('state'),
      $container->get('entity_type.manager'),
      $container->get('stream_wrapper_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function query() {
    $query = $this->select('file_managed', 'msfc')
      ->distinct()
      ->condition('msfc.status', TRUE)
      ->fields('msfc', ['type']);
    $query->addExpression("(SUBSTRING(msfc.uri, 1, INSTR(msfc.uri, '://')-1))", 'scheme');

    // Always filter by public streams, optionally restrict to a specific one.
    if (!empty($this->configuration['scheme'])) {
      $query->where("(SUBSTRING(msfc.uri, 1, INSTR(msfc.uri, '://')-1) = :scheme)", [
        ':scheme' => $this->configuration['scheme'],
      ]);
    }
    else {
      $query->where("(SUBSTRING(msfc.uri, 1, INSTR(msfc.uri, '://')-1) IN (:schemes[]))", [
        ':schemes[]' => $this->publicStreams,
      ]);
    }

    // Filter by source type if available.
    if (!empty($this->configuration['type'])) {
      if (!is_string($this->configuration['type'])) {
        throw new \LogicException('File entity source type must be the ID of a source file entity bundle');
      }
      $query->condition('msfc.type', $this->configuration['type']);
    }
    else {
      $query->condition('msfc.type', 'undefined', '<>');
    }

    return $query;
  }

  /**
   * {@inheritdoc}
   */
  public function fields() {
    return [
      'field_name' => $this->t('The field name of the corresponding media source field.'),
      'type' => $this->t('The entity bundle.'),
      'scheme' => $this->t('The scheme the media assets were stored with.'),
      'file_extensions' => $this->t('The allowed extensions of a media source field which base type is file.'),
      'alt_field' => $this->t('The alt field of an image source field is enabled.'),
      'alt_field_required' => $this->t('The alt field of an image source field is required.'),
      'title_field' => $this->t('The title field of an image source field is enabled.'),
      'title_field_required' => $this->t('The title field of an image source field is required.'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getIds() {
    return [
      'type' => [
        'type' => 'string',
        'alias' => 'msfc',
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function prepareRow(Row $row) {
    [
      'type' => $source_type,
      'scheme' => $scheme,
    ] = $row->getSource();

    if (!$source_type || !$scheme) {
      return FALSE;
    }

    // We have to find the media type whose source field's URI scheme matches
    // the current one.
    $media_type = $this->mediaTypeStorage->load($source_type);

    if ($media_type === NULL) {
      return FALSE;
    }

    assert($media_type instanceof MediaTypeInterface);
    $media_source = $media_type->getSource();
    $media_source_field_config_def = $media_source->getSourceFieldDefinition($media_type);
    $media_source_field_storage_def = $media_source_field_config_def->getFieldStorageDefinition();
    $additional_properties = [
      'field_name' => $media_source->getConfiguration()['source_field'],
    ];

    // If the media source plugin is an instance of the File plugin, that means
    // that the corresponding field configuration has an allowed file extensions
    // configuration. This is true for even AudioFile, VideoFile and Image.
    if ($media_source instanceof File) {
      assert($media_source_field_storage_def->getSetting('uri_scheme') === $scheme);
      // The field storage's URI scheme has to match the current scheme when
      // we have only one scheme.
      // @todo handle "max_filesize" as well.
      $additional_properties['file_extensions'] = $media_source_field_config_def->getSettings()['file_extensions'] ?? '';

      // We have to collect every different file extension of the existing
      // files.
      $required_extensions_results = $this->getDatabase()->select('file_managed', 'fmext')
        ->distinct()
        ->condition('fmext.status', TRUE)
        ->fields('fmext', ['uri'])
        ->condition('fmext.type', $source_type)
        ->where("SUBSTRING(fmext.uri, 1, INSTR(fmext.uri, '://')-1) = :scheme", [':scheme' => $scheme])
        ->execute()
        ->fetchAll();
      $required_extensions = array_reduce($required_extensions_results, function (array $extensions, $item) {
        $extension = pathinfo($item->uri)['extension'] ?? NULL;
        if ($extension && !in_array($extension, $extensions)) {
          $extensions[] = $extension;
        }
        return $extensions;
      }, []);

      if (!empty($required_extensions)) {
        $preexisting_extensions = !empty($additional_properties['file_extensions'])
          ? array_filter(array_unique(explode(' ', $additional_properties['file_extensions'])))
          : [];
        $additional_properties['file_extensions'] = implode(' ', array_unique(array_merge($preexisting_extensions, $required_extensions)));
      }
    }

    if ($media_source instanceof Image) {
      // Alt and title properties of an image field always exist, regardless
      // of whether they are accessible and editable on the entity edit form or
      // not.
      // But we also want to make sure that if either the alt or title property
      // was editable on the source site, then it will be editable also on the
      // destination site.
      // For first, we try to determine which properties should be enabled (and
      // thus accessible and editable on UI) based on the source media image
      // entities' alt and title fields' content. We assume that if there is at
      // least a single value for an image's title property in the source
      // database, then we have to change visibility of the title property to
      // TRUE, so that it can be edited on the destination site as well.
      $alt_title_config = $this->getImageAltTitleSettingsFromPropertyFieldContent();
      // All image field content is also migrated into a media entity. If any of
      // the alt or title properties shouldn't be necessarily shown (and be
      // editable) on the entity edit form based on the source media entity's
      // alt or title field value, we still have to check the configuration of
      // the source site's image fields.
      // If any of the preexisting image field was configured to show the alt or
      // the title property, then we will make their input field visible.
      if (!empty(array_diff(array_values($this->fieldNamePropertyMap), array_keys($alt_title_config)))) {
        $alt_title_config += $this->getSettingsFromImageFields();
      }

      // Get the 'required' settings from the image fields we found.
      $this->mergePropertyRequiredSettingsFromImageFields($alt_title_config);

      // Add the discovered and the default configuration.
      $additional_properties = $alt_title_config + [
        'alt_field' => TRUE,
        'alt_field_required' => TRUE,
        'title_field' => FALSE,
        'title_field_required' => FALSE,
      ] + $additional_properties;
    }

    foreach ($additional_properties as $property_name => $property_value) {
      $row->setSourceProperty($property_name, $property_value);
    }

    return parent::prepareRow($row);
  }

  /**
   * Discovers the image field settings based on existing property values.
   *
   * The alt and title properties of an image field always exist, regardless
   * of whether they are actually accessible and editable on the entity edit
   * form or not. This method checks the content of the corresponding alt and
   * title fields. If we find at least a single, non-empty row, we say that the
   * actual property should be shown on the destination media entity's edit
   * form.
   *
   * @return true[]
   *   An array of those field instance settings that should be revealed, keyed
   *   by the settings key ('alt_field', 'title_field').
   *   For example, if we have rows for alt, but don't have any data for title,
   *   the array returned will be this:
   *   @code
   *   [
   *     'alt_field' => TRUE
   *   ]
   *   @endcode
   */
  protected function getImageAltTitleSettingsFromPropertyFieldContent() {
    $data = [];

    foreach ($this->fieldNamePropertyMap as $field_name => $property) {
      $property_values_query = $this->getDatabase()->select("field_data_$field_name", $field_name)
        ->fields($field_name)
        ->condition("$field_name.{$field_name}_value", '', '<>')
        ->isNotNull("$field_name.{$field_name}_value");
      $property_values_present = (int) $property_values_query->countQuery()->execute()->fetchField() > 0;

      if ($property_values_present) {
        $data[$property] = TRUE;
      }
    }

    return $data;
  }

  /**
   * Discovers enabled properties based on the image field configurations.
   *
   * The alt and title properties of an image field always exist, regardless
   * of that they are actually accessible and editable on the entity edit form.
   * This method checks the config of every image field's instance configuration
   * When alt (or title) was enabled for at least one image field, we say that
   * the property should be shown on the destination media entity's edit
   * form.
   *
   * @return array
   *   An array of those settings that should be revealed, keyed by the settings
   *   key ('alt_field', 'title_field').
   */
  protected function getSettingsFromImageFields() {
    $data = [];
    $image_field_names = $this->getImageFieldNames();

    foreach ($image_field_names as $image_field_name) {
      $field_instance_config_results = $this->getDatabase()->select('field_config_instance', 'fci')
        ->fields('fci', ['data'])
        ->condition('fci.field_name', $image_field_name)
        ->condition('fci.entity_type', 'file', '<>')
        ->execute()
        ->fetchAll();

      foreach ($field_instance_config_results as $field_instance_config_result) {
        $field_config_data = unserialize($field_instance_config_result->data);

        foreach ($this->fieldNamePropertyMap as $property) {
          if (isset($field_config_data['settings'][$property]) && !empty($field_config_data['settings'][$property])) {
            $data[$property] = TRUE;
          }
        }

        if (empty(array_diff(array_values($this->fieldNamePropertyMap), array_keys($data)))) {
          break 2;
        }
      }
    }

    return $data;
  }

  /**
   * Merges alt/title 'required' settings based on image field discovery.
   *
   * If any image field have alt (or title) set up as optional, we don't let
   * them being required.
   *
   * @param array $data
   *   The discovered data about alt and title revealment.
   */
  protected function mergePropertyRequiredSettingsFromImageFields(array &$data) {
    foreach ($this->fieldNamePropertyMap as $field_name => $property) {
      $property_field_config_result = $this->getDatabase()->select('field_config_instance', 'fci')
        ->fields('fci', ['data'])
        ->condition('fci.field_name', $field_name)
        ->condition('fci.bundle', 'image')
        ->condition('fci.entity_type', 'file')
        ->execute()
        ->fetchAll();

      assert(count($property_field_config_result) === 1);

      if ($property_field_config_data = unserialize($property_field_config_result[0]->data)) {
        if (isset($property_field_config_data['required'])) {
          $not_set_or_not_required = !isset($data["{$property}_required"]) || empty($data["{$property}_required"]);
          $data["{$property}_required"] = $not_set_or_not_required && !empty($data[$property]) && !empty($property_field_config_data['required']);
        }
      }
    }
  }

}
