<?php

namespace Drupal\Tests\media_migration\Kernel\Migrate;

/**
 * Tests media migration.
 *
 * @group media_migration
 */
class MediaMigrationTest extends MediaMigrationTestBase {

  /**
   * Tests the migration of media entities.
   *
   * @dataProvider providerTestMediaMigration
   */
  public function testMediaMigration(string $destination_token, string $reference_method, bool $classic_node_migration, array $expected_node1_embed_attributes) {
    $this->setEmbedTokenDestinationFilterPlugin($destination_token);
    $this->setEmbedMediaReferenceMethod($reference_method);
    $this->setClassicNodeMigration($classic_node_migration);

    // The Drupal 8|9 entity revision migration causes a file not found
    // exception without properly migrated files. For this test, it is enough to
    // properly migrate the public files.
    $fs_fixture_path = implode(DIRECTORY_SEPARATOR, [
      DRUPAL_ROOT,
      drupal_get_path('module', 'media_migration'),
      'tests',
      'fixtures',
    ]);
    $file_migration = $this->getMigration('d7_file');
    $source = $file_migration->getSourceConfiguration();
    $source['constants']['source_base_path'] = $fs_fixture_path;
    $file_migration->set('source', $source);

    $this->executeMigration($file_migration);
    $this->executeMigrations([
      'd7_view_modes',
      'd7_field',
      'd7_comment_type',
      'd7_node_type',
      'd7_field_instance',
      'd7_media_source_field_config',
    ]);
    // Abusing ::startCollectingMessages(): "d7_field_formatter_settings"
    // triggers PluginNotFoundException: The "file_rendered" plugin does not
    // exist.
    $this->startCollectingMessages();
    $this->executeMigrations([
      'd7_field_formatter_settings',
    ]);
    $this->stopCollectingMessages();
    $this->executeMigrations([
      'd7_field_instance_widget_settings',
      'd7_embed_button_media',
      'd7_filter_format',
      // @todo: Every migration that uses the "media_wysiwyg_filter" process
      // plugin should depend on "d7_media_view_modes".
      'd7_media_view_modes',
      // Nodes and media entities need an owner.
      'd7_user_role',
      'd7_user',
      $classic_node_migration ? 'd7_node' : 'd7_node_complete',
      'd7_file_entity_type',
      'd7_file_entity',
    ]);

    // Check configurations.
    $this->assertArticleImageFieldsAllowedTypes();
    $this->assertArticleMediaFieldsAllowedTypes();

    // Check the migrated media entities.
    $this->assertMedia1FieldValues();
    $this->assertMedia2FieldValues();
    $this->assertMedia3FieldValues();
    $this->assertMedia6FieldValues();
    $this->assertMedia7FieldValues();
    $this->assertMedia8FieldValues();
    $this->assertMedia9FieldValues();
    $this->assertMedia10FieldValues();
    $this->assertMedia11FieldValues();
    $this->assertMedia12FieldValues();

    $this->assertNode1FieldValues($expected_node1_embed_attributes);
  }

  /**
   * Data provider for ::testMediaTokenToMediaEmbedTransform().
   *
   * @return array
   *   The test cases.
   */
  public function providerTestMediaMigration() {
    $default_attributes = [
      'data-entity-type' => 'media',
      'alt' => 'Different alternative text about blue.png in the test article',
      'title' => 'Different title copy for blue.png in the test article',
      'data-align' => 'center',
    ];

    $test_cases = [
      // ID reference method. This should be neutral for media_embed token
      // transform destination.
      'Entity embed destination, ID reference method, classic node migration' => [
        'Destination filter' => 'entity_embed',
        'Reference method' => 'id',
        'Classic node migration' => TRUE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-id' => '1',
            'data-embed-button' => 'media',
            'data-entity-embed-display' => 'view_mode:media.wysiwyg',
          ] + $default_attributes,
        ],
      ],
      'Media embed destination, ID reference method, classic node migration' => [
        'Destination filter' => 'media_embed',
        'Reference method' => 'id',
        'Classic node migration' => TRUE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-uuid' => TRUE,
            'data-view-mode' => 'wysiwyg',
          ] + $default_attributes,
        ],
      ],
      'Entity embed destination, ID reference method, complete node migration' => [
        'Destination filter' => 'entity_embed',
        'Reference method' => 'id',
        'Classic node migration' => FALSE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-id' => '1',
            'data-embed-button' => 'media',
            'data-entity-embed-display' => 'view_mode:media.wysiwyg',
          ] + $default_attributes,
        ],
      ],
      'Media embed destination, ID reference method, complete node migration' => [
        'Destination filter' => 'media_embed',
        'Reference method' => 'id',
        'Classic node migration' => FALSE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-uuid' => TRUE,
            'data-view-mode' => 'wysiwyg',
          ] + $default_attributes,
        ],
      ],
      // UUID reference method.
      'Entity embed destination, UUID reference method, classic node migration' => [
        'Destination filter' => 'entity_embed',
        'Reference method' => 'uuid',
        'Classic node migration' => TRUE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-uuid' => TRUE,
            'data-embed-button' => 'media',
            'data-entity-embed-display' => 'view_mode:media.wysiwyg',
          ] + $default_attributes,
        ],
      ],
      'Media embed destination, UUID reference method, classic node migration' => [
        'Destination filter' => 'media_embed',
        'Reference method' => 'uuid',
        'Classic node migration' => TRUE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-uuid' => TRUE,
            'data-view-mode' => 'wysiwyg',
          ] + $default_attributes,
        ],
      ],
      'Entity embed destination, UUID reference method, complete node migration' => [
        'Destination filter' => 'entity_embed',
        'Reference method' => 'uuid',
        'Classic node migration' => FALSE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-uuid' => TRUE,
            'data-embed-button' => 'media',
            'data-entity-embed-display' => 'view_mode:media.wysiwyg',
          ] + $default_attributes,
        ],
      ],
      'Media embed destination, UUID reference method, complete node migration' => [
        'Destination filter' => 'media_embed',
        'Reference method' => 'uuid',
        'Classic node migration' => FALSE,
        'expected_node1_embed_html_attributes' => [
          0 => [
            'data-entity-uuid' => TRUE,
            'data-view-mode' => 'wysiwyg',
          ] + $default_attributes,
        ],
      ],
    ];

    // Drupal 8.8.x only has 'classic' node migrations.
    // @see https://www.drupal.org/node/3105503
    if (version_compare(\Drupal::VERSION, '8.9', '<')) {
      $test_cases = array_filter($test_cases, function ($test_case) {
        return $test_case['Classic node migration'];
      });
    }

    return $test_cases;
  }

}
